#!/usr/bin/perl
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# The icons used for the up and back images are modified from the Nuovo icon
# set for the Gnome desktop.


use strict;
use Encode::Unicode;
use Tk;
use Tk::JPEG;
use Tk::PNG;
use Tk::TextUndo;
use Tk::ROText;
use Crypt::Blowfish;
use Crypt::CBC;
use IO::Compress::Gzip qw(gzip $GzipError) ;
use IO::Uncompress::Gunzip qw(gunzip $GunzipError) ;
use Date::Manip;
use Data::Dumper;
use LWP::Simple;
use MIME::Base64;
#use Tk::FontDialog;
use Tk::DateEntry;
use Tk::Dialog;
use Tk::DialogBox;
use File::Temp qw/ tempfile tempdir /;
use File::Copy;

$main::DEBUG=0;
$main::ostype=$^O;
$main::windows_os=0;
$main::windows_os =1 if($main::ostype =~ m/Win/);

if($main::windows_os)
  {
    $main::ENV{'TZ'}='EST' 
      unless(exists $main::ENV{'TZ'}); # kludge for a Windows problem
  }
else
  {
    require Image::Magick;
    require Text::Aspell;
  }

use constant EM_DASH => "\x{2014}";
use constant LEFT_SINGLE_QUOTE => "\x{2018}";
use constant RIGHT_SINGLE_QUOTE => "\x{2019}";
use constant LEFT_DOUBLE_QUOTE => "\x{201c}";
use constant RIGHT_DOUBLE_QUOTE => "\x{2019}";

$main::tempfilehand='';
$main::tempfilename='';

$main::tmpwritten=0;
$main::dbopened=0;

if($main::windows_os)
  {
    $main::options_file = './dream-options.txt';
  }
else
  {
    my $home=$main::ENV{'HOME'};
    $main::options_file = "$home/.diarydb";
  }
# read configuration
if(-e $main::options_file)
  {
    do $main::options_file;
  }

unless(defined $main::dream_options{'quesbg'})
  {
    reset_options_ram();
    save_options();
  }


# Create a window 
$main::toplevel = MainWindow -> new();
$main::toplevel->title('Dream Diary');
$main::toplevel->protocol('WM_DELETE_WINDOW',\&quit_button); 

#$main::tkf = $main::toplevel -> FontDialog();
#print $main::toplevel->GetDescriptiveFontName($main::tkf -> Show());

$main::dscale = $main::toplevel -> Scale(
		      -length => 800,
		      -showvalue => 1,
		      -from => 0,
		      -orient => 'horizontal',
		      -troughcolor=>$main::dream_options{'quesbg'});
$main::entered_date='';

$main::date_entry = $main::toplevel 
      ->DateEntry(-font => $main::dream_options{'date-font'},
		  -textvariable => \$main::entered_date,
		  -parsecmd => \&parse_calendar,
		  -formatcmd => \&format_calendar,
		  -state=> 'readonly',
		  -configcmd 
		  => sub
		  { 
		      my (%args) = @_;
		      return unless ($args{-date});
		    return unless(@{$args{-date} });
		    my($day,$month,$year) = @{$args{-date} };
		    
		    my $canonical_date = 
		      UnixDate(ParseDate("$month/$day/$year"),'%Y%m%d');
		    my $widget = $args{-widget};
		    my $datewidget = $args{-datewidget};
		    if(exists $main::dreams_bydate{$canonical_date})
		      {
			my $numeric = 
			  $main::dreams_bydate{$canonical_date} -> [2];
			$datewidget->configure(
					       -bg 
					       => $main::dream_options{'dreambg'}
					       ->[$numeric],
					       -fg => $main::dream_options{'dreamfg'}
					       ->[$numeric],);
		      }
		    else
		      {
			$datewidget->configure(
					       -bg 
					       => $main::dream_options{'bgcolor'},
					       -fg => $main::dream_options{'fgcolor'},
					      );
			
		      };
		   }
		 );
$main::date_entry -> Subwidget('entry') 
  -> configure(-width=>0,
	     -justify=>'right');
unless(exists $main::dream_options{'up-arrow'})
  {
    reset_arrows_ram();
    save_options();
  }
$main::up_image = $main::toplevel -> 
  Photo(-data=>$main::dream_options{'up-arrow'});
$main::back_image = $main::toplevel -> 
  Photo(-data=>$main::dream_options{'back-arrow'});

		  
$main::back_button=$main::toplevel->
  Button(
	 -bg => $main::dream_options{'bgcolor'},
	 -fg => $main::dream_options{'fgcolor'},
	 -state=>'disabled',
	 -text=>'',
	 -font => $main::dream_options{'date-font'},
	 -command => \&undo_search
	);


$main::dream_numeric=5;

@main::dream_description_array=('',
				'Joyful',
				'Transcendent',
				'Past Life',
				'Prophetic',
				'Average',
				'Lucid',
				'Intense',
				'Nightmare',
				'Epic'
			       );


						
$main::paned_window = $main::toplevel -> Panedwindow(-orient=>'vertical',);

$main::dream_text = $main::toplevel ->
    Scrolled('TextUndo',
	     -fg => $main::dream_options{'dreamfg'}->[5],
	     -wrap => 'word',
	     -font => $main::dream_options{'main-font'},
	     -height => 10,
	     -spacing3=>'2m',
	     -scrollbars => 'w',
	     -exportselection => 1);
$main::dream_text -> Subwidget('yscrollbar')
-> configure( -bg => $main::dream_options{'bgcolor'},
	      -troughcolor=>$main::dream_options{'quesbg'});

$main::dream_text ->
    tagConfigure('ital',
		 -font =>$main::dream_options{'ital-font'},
		 );

$main::dream_text ->
    tagConfigure('bold',
		 -font => $main::dream_options{'quote-font'});

$main::dream_text ->
    tagConfigure('underline',
		 -underline =>1);

$main::dream_interp = $main::toplevel ->
    Scrolled('TextUndo',
	     -font =>
	     $main::dream_options{'main-font'},
	     -wrap => 'word',
	     -height => 10,
	     -spacing3=>'2m',
	     -scrollbars => 'w',
	     -exportselection => 1);

$main::dream_interp ->
    tagConfigure('ital',
		 -font =>
		 $main::dream_options{'ital-font'},);
$main::dream_interp ->
    tagConfigure('quotecolor',
		 -foreground =>
		 'RoyalBlue',
		 -underline => 1,
		 -font => $main::dream_options{'quote-font'});

$main::dream_interp ->
    tagConfigure('bold',
		 -font => $main::dream_options{'quote-font'});

$main::dream_interp ->
    tagConfigure('underline',
		 -underline =>1);

$main::dream_interp ->
    tagConfigure('active_date',
		 -foreground =>
		 'ForestGreen',
		 -underline=>1,
		 -font => $main::dream_options{'active-date-font'});

$main::dream_interp ->
    tagBind('active_date',
	    '<3>',
	    \&active_date
	    );

$main::dream_interp ->
    tagBind('quotecolor',
	    '<3>',
	    \&active_quote
	    );


$main::active_text_widget = 0;

%main::dreaminfo = (
		    'file-name'=>'',
		    'number_ref' => 0,
		    'date' => '',
		    'dream_scale' => $main::dscale,
		    'entry-changed' => 0,
		    'active-text' => \$main::active_text_widget,
		    'db-dirty'=>0,
		    'passphrase' => '',
	      );

$main::dream_text -> bind("<Any-KeyPress>"=> 
			  [sub{
			     ${$main::dreaminfo{'active-text'}}=shift;
			     unless(%main::dreams_bydate)
			       {
				 my $char = shift;
				 new_database();
				 $main::dream_text->
				   insert('end',$char)
				     unless(length($char)>1);
				 }
			     $main::dreaminfo{'entry-changed'} = 1;
			     $main::dreaminfo{'db-dirty'}=1;
			   },Ev('K')]);

$main::dream_text -> bind('<ButtonPress>'=>
		    sub{
		      ${$main::dreaminfo{'active-text'}}=shift;
		    }
		  );

$main::dream_interp -> bind("<Any-KeyPress>" => 
		      [sub{
			${$main::dreaminfo{'active-text'}}=shift;
			 unless(%main::dreams_bydate)
			       {
				 my $char = shift;
				 new_database();
				 $main::dream_interp->
				   insert('end',$char)
				     unless(length($char)>1);
				 }
			$main::dreaminfo{'entry-changed'} = 1;
			$main::dreaminfo{'db-dirty'}=1;
			
		      },Ev('K')]);

$main::dream_interp -> bind('<ButtonPress>'=>
		      sub{
			${$main::dreaminfo{'active-text'}}=shift;
		      }
		     );
		      
  
  
  $main::todays_date=UnixDate(ParseDate('today'),'%Y%m%d');


@main::dates_inorder=();

$main::menubar='';
$main::dream_intensity='';  # set in setup_menus

&setup_menus(
	     \$main::menubar,
	   );

post_colors();
$main::paned_window->add($main::dream_text);
$main::paned_window->add($main::dream_interp);
$main::menubar->grid('-','-',-sticky=>"ew");
$main::date_entry->grid($main::back_button,$main::dream_intensity);
$main::dscale->grid('-','-',-sticky=>'ew');
$main::paned_window->grid('-','-',-sticky=>'news');



$main::toplevel->gridRowconfigure(0,-weight=>0,);
$main::toplevel->gridRowconfigure(1,-weight=>0,);
$main::toplevel->gridRowconfigure(2,-weight=>0,);
$main::toplevel->gridRowconfigure(3,-weight=>1,);

$main::toplevel->gridColumnconfigure(0,-weight=>1);
$main::toplevel->gridColumnconfigure(1,-weight=>1);
$main::toplevel->gridColumnconfigure(2,-weight=>1);


#################################################

MainLoop();

sub parse_calendar
  {
    my ($date_selected)=@_;
     my $pdate = ParseDate($date_selected);
    $pdate=~m/\A(\d\d\d\d)(\d\d)(\d\d)/;
    my $year=$1;
    my $month=$2;
    my $day=$3;
    $day=~s/\A0//;
    die "Bad $year,$month,$day\n" unless($pdate);
    return unless($pdate);
    my $canonical_date = UnixDate($pdate,'%Y%m%d');
    access_dream($canonical_date);
    return ($year,$month,$day);
  }

sub format_calendar
  {
    my($year,$month,$day)=@_;
    $day=~s/\ /0/;
    $month=~s/\ /0/;
    my $pdate = ParseDate("$month/$day/$year");
    die "Bad $year,$month,$day\n" unless($pdate);
    my $canonical_date = UnixDate($pdate,'%Y%m%d');
    access_dream($canonical_date);
    $main::entered_date = '';
    return UnixDate($canonical_date,'%A, %b %e %Y');
  }


sub post_colors
  {
    $main::toplevel -> 
      configure( -bg => $main::dream_options{'bgcolor'},
		 -fg => $main::dream_options{'fgcolor'});
    $main::paned_window -> 
      configure( -bg => $main::dream_options{'bgcolor'});
    $main::dream_intensity -> 
      configure( -bg => $main::dream_options{'bgcolor'},
		 -fg => $main::dream_options{'fgcolor'});
    $main::dreaminfo{'dream_scale'} -> 
      configure( -bg => $main::dream_options{'bgcolor'},
		 -fg => $main::dream_options{'fgcolor'},);
    $main::dream_intensity -> 
      configure( -bg => $main::dream_options{'bgcolor'},
		 -fg => $main::dream_options{'fgcolor'},);

    $main::dream_text -> Subwidget('yscrollbar')
      -> configure( -bg => $main::dream_options{'bgcolor'},
		    -troughcolor=>$main::dream_options{'quesbg'});

    $main::dream_interp -> 
      configure( -bg => $main::dream_options{'interpbg'},
		 -fg => $main::dream_options{'interpfg'},);
    $main::dream_interp -> Subwidget('yscrollbar')
      -> configure( -bg => $main::dream_options{'bgcolor'},
		    -troughcolor=>$main::dream_options{'quesbg'});
    $main::menubar->
      configure( -bg => $main::dream_options{'bgcolor'},
		 -fg => $main::dream_options{'fgcolor'},);
    $main::date_entry -> configure(
				   -bg => $main::dream_options{'bgcolor'},
				   -fg => $main::dream_options{'fgcolor'},
				   -boxbackground=> $main::dream_options{'bgcolor'});
    $main::date_entry -> Subwidget('entry')
      ->configure( -disabledbackground => $main::dream_options{'bgcolor'},
		   -disabledforeground => $main::dream_options{'fgcolor'},);
    foreach my $child ($main::menubar-> children())
      {
	Tk::catch
	    {
	      $child -> configure(-bg => $main::dream_options{'bgcolor'},
				  -fg => $main::dream_options{'fgcolor'},
				  );
	    };
	foreach my $child2 ($child-> children())
	  {
	    $child2 -> configure(-bg => $main::dream_options{'bgcolor'},
				 -fg => $main::dream_options{'fgcolor'});
	  }
      }
  }

sub close_out_db
{
    my $dbname=$main::dreaminfo{'file-name'};
    if ($main::dreaminfo{'db-dirty'})
    {
	my $dlg = $main::toplevel->Dialog(
	    -title          => 'Save changes?',
	    -text           => 'Database changed since save.',
	    -default_button => 'Save',
	    -buttons        => ['Save','Discard','Cancel'],
	    -bg => $main::dream_options{'bgcolor'},
	    -fg => $main::dream_options{'fgcolor'}
	    );
	my $response = $dlg -> Show(-global);
	return 1 if ($response eq 'Cancel');
	if($response eq 'Discard')
	{
	    rename "$dbname.old",$dbname;
	    $main::tmpwritten=0;
	    return 0;
	}
	save_button() if ($response eq 'Save');
	copy($main::tempfilename,$dbname);
    }
    else # no edits since last save
    {
	if($main::tmpwritten)
	{
	    copy($main::tempfilename,$dbname);    
	}
	else # nothing was ever saved anywhere
	{
	    rename "$dbname.old",$dbname;
	}
    }
    $main::dbopened=0;
    $main::tmpwritten=0;
    return 0;
}

sub quit_button
  {
# if the user canceled the close out, it returns 1
      return if (close_out_db());
      exit();
  }

sub new_database
  {
      if($main::dbopened)
      {
# if the user canceled the close out, it returns 1
	  return if(close_out_db());
      };
      $main::dbopened=1;
    $main::dreaminfo{'passphrase'}=
	new_passphrase(\$main::dreaminfo{'passphrase'});
    return unless($main::dreaminfo{'passphrase'});
    %main::dreams_bydate = ($main::todays_date => [
				   '',   #dream
				   '',   #interp
				   5]); 
    create_dates(\@main::dates_inorder);
    @main::date_stack = ();
    push @main::date_stack,({'range'=>\@main::dates_inorder,
			     'place'=>-1});
    
    $main::dreaminfo{'date'}=$main::todays_date;
    $main::dscale -> configure(
			      -to => 0,
			      -tickinterval=>1,
			      -variable => 
			      \{$main::date_stack[0]->{'place'}},
			     );

    $main::dreaminfo{'entry-changed'} = 0;
    $main::dreaminfo{'db-dirty'}=1;
    $main::dreaminfo{'file-name'} = '';
    post_pending_changes();
  }

sub open_database
  {
    my (	
	$filename,
       )=@_;
    my $tempname;
    if($main::dreaminfo{'db-dirty'})
      {
	my $response = communicate('Do you want to save the current database?',
			       ['Yes','No','Cancel']);
	return if ($response eq 'Cancel');
	save_button('') if ($response eq 'Yes');
      }
    if($filename)
      {
	$tempname = $filename;
      }
    else
      {
	$tempname=communicate_filename
	  (
	   'Select the database to open'
	  );
      }
    $main::dreaminfo{'file-name'}=$tempname;
    $main::toplevel->Busy(-recurse=>1);
    open DFILE,"<$tempname" or 
	open DFILE,"<$tempname.old" or 
	die "unable to open input file\n";
    binmode DFILE;
    my $buffer;
    my $bytes_read;
    my $db_in='';
    while ($bytes_read = read(DFILE,$buffer,10000) !=0)
      {
	$db_in .= $buffer;
      }
    print "input size =",length($db_in),"\n" if $main::DEBUG;
    close DFILE;

    $main::dreaminfo{'passphrase'} = decrypt_file($db_in);
# if  $main::dreaminfo{'passphrase'} is empty the user
# must have canceled opening the database
    unless ($main::dreaminfo{'passphrase'})
    {
	$main::toplevel->Unbusy();
	return;
    }
    rename $tempname, "$tempname.old"; 
    $main::dbopened=1;
    create_dates(\@main::dates_inorder);
    my $dream_number = @main::dates_inorder;
    $main::toplevel->Unbusy();
    push @main::date_stack,({'range'=>\@main::dates_inorder,
			     'place'=>$dream_number});
    
    my $ndreams = @main::dates_inorder;
    
    $main::dscale -> configure(-command => [
					   \&post_pending_changes],
			      -to => $ndreams-1,
			      -tickinterval=>set_ticks($ndreams),
			      -variable => \$main::date_stack[0]->{'place'},
			     );
    post_pending_changes();
  }

sub setup_menus()
  {
    my(
       $menubar_ref,
      )=@_;


    $main::dream_intensity = $main::toplevel 
      -> Menubutton(-text => 'Dream properties',
		    -font => $main::dream_options{'date-font'});
    my $command_string = '';
    my $j=-1;

# A Perl "macro":
# We need to do this to get actual values for $j plugged into the
# procedures, rather than $j itself (which will be meaningless
# when the procedures are asynchronously triggered.
    foreach my $desc (@main::dream_description_array)
      {
	$j++;
	next unless($desc =~ m/\S/);
	$command_string .=
	"\$main::dream_intensity 
	  -> AddItems([\'command\'=> \$main::dream_description_array[$j],
		       -bg => \$main::dream_options{\'dreamfg\'} -> [$j],
		       -foreground => \$main::dream_options{\'dreambg\'} -> [$j],
		       -command => sub 
		       {
			 \$main::dream_numeric=$j;
			 \$main::dreaminfo{\'entry-changed\'}=1;
                         \$main::dreaminfo{\'db-dirty\'}=1;
                         \$main::dream_text
                            ->configure(-bg =>\$main::dream_options{\'dreambg\'} -> [$j]);
                         \$main::dream_text
                            ->configure(-fg =>\$main::dream_options{\'dreamfg\'} -> [$j]);
		        \$main::dream_intensity
                             ->configure(
		          -text=>\$main::dream_description_array[$j],
                          -fg => \$main::dream_options{\'dreamfg\'} -> [$j],
                          -bg => \$main::dream_options{\'dreambg\'} -> [$j]);}],
		     );\n";
      }
    eval($command_string);


    my @menus;

    $$menubar_ref = $main::toplevel
      ->Frame(-relief=>'ridge',-borderwidth=>2);
    foreach (qw/File Edit Search Images Config Help/)
      {
	push (@menus,$$menubar_ref->Menubutton(-text => $_,));
      }



   
    $menus[0]->pack(-side=>'left');
    $menus[1]->pack(-side=>'left');
    $menus[2]->pack(-side=>'left');
    $menus[3]->pack(-side=>'left');
    $menus[4]->pack(-side=>'left');
    $menus[5]->pack(-side=>'left');

    
    $menus[0]-> AddItems(['command'=>'New Database',
			  -command=>\&new_database],
			 ['command'=>'Open Database',
			  -command=>[\&open_database,
				     '']],);
    my $i=0;
    foreach my $file (@{$main::dream_options{'file-history'}})
      {
	$menus[0]-> AddItems(['command'=>"Open $file",
			      -command=>[\&open_database,
					 $file]],
			    );
      }
    
    $menus[0]-> AddItems(
			 ['command'=>'Save',
			  -accelerator=>'Ctrl-S',
			  -command=>[\&save_button,
				     '']],
			 ['command'=>'Quit',
			  -accelerator => 'Ctrl-Q', 
			  -command=>[\&quit_button,
				     '']]);
    
    $main::toplevel->bind('Tk::TextUndo','<Control-s>',
	     sub{});
    
    $main::toplevel ->bind('<Control-s>',\&save_button);
    
    $main::toplevel->bind('Tk::TextUndo','<Control-q>',
	     sub{});
    
    $main::toplevel ->bind('<Control-q>',\&quit_button);
    
    $main::toplevel->bind('Tk::TextUndo','<Control-n>',
	     sub{});
    
    $main::toplevel->bind('Tk::TextUndo','<Control-i>',
	     sub{});

 $main::toplevel->bind('Tk::TextUndo','<Control-b>',
	     sub{});

    
    $main::toplevel ->bind('<Control-n>',[\&access_dream_key,
			     $main::todays_date]);
    
    $main::toplevel->bind('Tk::TextUndo','<Control-g>',
	     sub{});

    $main::toplevel ->bind('<Control-g>',\&get_date_goto_dream,  # omitting 
			     # the $m widget from
			     # the calling sequence because
			     # it is implied
);

    
    $main::dream_text -> bind('<Control-i>',[\&style_text,
				       'ital']);
    
    $main::dream_interp -> bind('<Control-i>',[\&style_text,
					 'ital']);

    $main::dream_text -> bind('<Control-b>',[\&style_text,
				       'bold']);
    
    $main::dream_interp -> bind('<Control-b>',[\&style_text,
					 'bold']);

    $main::dream_text -> bind('<Control-u>',[\&style_text,
				       'underline']);
    
    $main::dream_interp -> bind('<Control-u>',[\&style_text,
					 'underline']);
    $main::dream_interp -> bind('<Alt-b>',[\&style_text,
				     'quotecolor']);
    
    $menus[1]->AddItems(
			['command' => 'Undo',
			 -accelerator => 'Ctrl-Z',
			 -command => [\&undo_key,
				     $main::dreaminfo{'active-text'}]],
			['command' => 'Cut',
			 -accelerator => 'Ctrl-X',
			 -command => [\&cut_key,
				     $main::dreaminfo{'active-text'}]],
			['command' => 'Copy',
			 -accelerator => 'Ctrl-C',
			 -command => [\&copy_key,
				     $main::dreaminfo{'active-text'}]],
			['command' => 'Paste',
			 -accelerator => 'Ctrl-V',
			 -command => [\&paste_key,
				     $main::dreaminfo{'active-text'}]],
			'-',
			['command' => 'New Dream', 
			 -accelerator =>'Ctrl-N',
			 -command=>[\&access_dream,
				    $main::todays_date]],
			
			'-',);
    

    $menus[1]->AddItems(
			['command'=>'Check Spelling',
			 -command=> [\&call_spelling,
				     $main::dream_text,
				     $main::dream_interp,
				    ]],
			'-',) unless($main::windows_os);
    
    $menus[1]->AddItems(
			['command'=>'Italics',
			 -accelerator=>'Ctrl-I',
			 -command => [\&style_text,
				      $main::dreaminfo{'active-text'},
				      'ital'
				     ]],
			['command'=>'Underline',
			 -accelerator=>'Ctrl-U',
			 -command => [\&style_text,
				      $main::dreaminfo{'active-text'},
				      'underline'
				     ]],
			['command'=>'Bold',
			 -accelerator=>'Ctrl-B',
			 -command => [\&style_text,
				      $main::dreaminfo{'active-text'},
				      'bold'
				     ]],
			'-',
			['command'=>'Bold (Active Quote)',
			 -accelerator=>'Alt-B',
			 -command=>[\&style_text,
				    $main::dreaminfo{'active-text'},
				    'quotecolor'
				   ]],
			['command'=>'Active Date',
			 -command=> [\&style_text,
				     $main::dreaminfo{'active-text'},
				     'active_date'
				    ]],
			'-',
			['command'=>'Delete current entry',
			 -command=> [\&delete_dream,
				     \%main::dreams_bydate,
				    ]]
		       );
  
    $menus[2]->AddItems(
			['command'=>'Go to date',
			 -accelerator=>'Ctrl-G',
			 -command=>[\&get_date_goto_dream,
				    '',]],
			['command'=>'Search on set of words',
			 -accelerator=>'Ctrl-W',
			 -command =>[\&rescan_button,
				     $main::toplevel
				    ]],
			['command'=>'Search on exact phrase',
			 -accelerator=>'Ctrl-F',
			 -command =>[\&search_exact_phrase_button,
				     $main::toplevel
				    ]],
			);

    $main::toplevel ->bind('<Control-w>',\&rescan_button,  # omitting 
			     # the $m widget from
			     # the calling sequence because
			     # it is implied.
			     );


    $main::toplevel ->bind('<Control-f>',\&search_exact_phrase_button,  
					  # omitting 
			     # the $m widget from
			     # the calling sequence because
			     # it is implied.
			     );

    
    $menus[3] -> AddItems(	
			   ['command'=>'Add image from file',
			    -command=> [\&add_image_from_file,
					$main::dreaminfo{'active-text'},
				       ]],
			   ['command'=>'Paste image from web',
			    -command=> [\&add_image_from_web,
					$main::dreaminfo{'active-text'},
				       ]]
			  );

    $menus[4]-> AddItems(['command'=>'Reset all options',
			  -command=>\&reset_all],
			 ['command'=>'Change Passphrase',
			  -command=>[\&new_passphrase,
				     \$main::dreaminfo{'passphrase'},
				     ]],
			 '-',
			 ['command'=>'Default color scheme',
			  -command=>\&default_color_scheme,
				    ],
			 ['command'=>'Change color scheme',
			  -command=>\&color_scheme,
				     ]
			);

    $menus[5]-> AddItems(['command'=>'About',
			  -command=>[\&communicate,
				    'Dream Diary']],
			 ['command'=>'Overview',
			  -command=>\&help_overview],
			 ['command'=>'Search Methods',
			  -command=>\&help_search],
			 ['command'=>'Inserting images',
			  -command=>\&help_insert_image],
			);


  }

sub default_color_scheme
  {
    reset_colors_ram();
    save_options();
    post_colors();
  }


sub reset_fonts_ram
  {
    $main::dream_options{'main-font'}= "{Bitstream Vera Sans} 16";
    $main::dream_options{'ital-font'}='{Bitstream Vera Sans} 16 italic';
    $main::dream_options{'quote-font'}='{Bitstream Vera Sans} 16 bold';
    $main::dream_options{'active-date-font'}='{Bitstream Vera Sans} 16';
    $main::dream_options{'date-font'} = '{Bitstream Vera Sans} 24';
    $main::dream_options{'symbol-font'} = '{standard symbols l} -16';
  }

sub reset_options_ram
  {
    
    $main::dream_options{'file-history'}=[];
    reset_fonts_ram();
    reset_colors_ram();
    reset_arrows_ram();
  }

sub reset_colors_ram
  {
    $main::dream_options{'bgcolor'}='MediumTurquoise';
    $main::dream_options{'fgcolor'}='White';
    $main::dream_options{'dreambg'}=[
				     'gray', # ''
				     'DarkOrchid2', # ''
				     'DarkSeaGreen1', # ''
				     'floral white', # 'Past Life'
				     'LightCyan1', # Prophetic
				     'AliceBlue', # 'Average'
				     'yellow', # 'Lucid',
				     'dark magenta',#'Intense'
				     'black',# 'Nightmare' 
				     'white'   # 'Epic'
				    ];
    $main::dream_options{'dreamfg'}=[
				     'gray', # ''
				     'yellow1', # ''
				     'turquoise4', # ''
				     'medium aquamarine', # 'Past Life'
				     'deep sky blue', # Prophetic
				     'navy', # 'Average'
				     'blue', # 'Lucid',
				     'orange',#'Intense'
				     'purple',# 'Nightmare' 
				     'red'   # 'Epic'
				    ];
    $main::dream_options{'interpbg'}='FloralWhite';
    $main::dream_options{'interpfg'}='black';
    $main::dream_options{'quesbg'}='FloralWhite';
    $main::dream_options{'quesfg'}='black';
  }

sub reset_arrows_ram
  {
    $main::dream_options{'up-arrow'}= 'iVBORw0KGgoAAAANSUhEUgAAACoAAAAyCAYAAAFdQXk7AAAABGdBTUEAAYagMeiWXwAACjNJ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';

$main::dream_options{'back-arrow'}= 'iVBORw0KGgoAAAANSUhEUgAAADIAAAAqCAYAAAGGvTjRAAAABGdBTUEAAYagMeiWXwAACLtJ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';
  }

sub reset_all
  {
    reset_options_ram();
    save_options();
    post_colors();
  }

sub undo_key
  {
    my($text_ref)=@_;
    return unless(ref($text_ref) eq 'REF');
    my $widget = $$text_ref;
    return unless(ref($widget) eq 'Tk::TextUndo');
    $widget -> eventGenerate('<Control-z>');
  }

sub cut_key
  {
    my($text_ref)=@_;
    return unless(ref($text_ref) eq 'REF');
    my $widget = $$text_ref;
    return unless(ref($widget) eq 'Tk::TextUndo');
    $widget -> eventGenerate('<Control-x>');
  }

sub copy_key
  {
    my($text_ref)=@_;
    return unless(ref($text_ref) eq 'REF');
    my $widget = $$text_ref;
    return unless(ref($widget) eq 'Tk::TextUndo');
    $widget -> eventGenerate('<Control-c>');
  }

sub paste_key
  {
    my($text_ref)=@_;
    return unless(ref($text_ref) eq 'REF');
    my $widget = $$text_ref;
    return unless(ref($widget) eq 'Tk::TextUndo');
    $widget -> eventGenerate('<Control-v>');
  }

sub add_image_from_file
  {
    my ($active_text_ref)=@_;
    my $pdata;
    my $imag;
    my $message='Select image file';
    my $types;
    if($main::windows_os)
    {
	$types=[ ['Jpeg files', '.jpg'],
		 ['Jpeg files', '.jpeg'],
		 ['PNG files',  '.png'],
	    ];}
    else
    {
	$types=[ ['Jpeg files', '.jpg'],
		 ['Jpeg files', '.jpeg'],
		 ['PNG files',  '.png'],
		 ['BMP files','.bmp'],
		 ['GIF files','.gif'],
		 ['SVG files','.svg'],
		 ['PS files','.ps'],
		 ['PS files','.eps'],
	    ];
    }
    while(1)
      {
	my $filed=communicate_filename($message,$types);
	return unless($filed);
        my $idata;	
        $message = 'Not an image file, re-select';
	if($main::windows_os)
	  {
	    open IFILE,"<$filed";
	    undef $/;
	    $pdata = <IFILE>;
	    $/="\n";
	    close IFILE;
	  }
	else
	  {
	    $imag = Image::Magick->new;
	    $idata = $imag -> Read($filed);
	  }
	last if($pdata);
	last unless($idata);
	
      }
    $pdata=($imag->ImageToBlob(magick=>'JPEG'))[0]
      unless($main::windows_os);

    post_image_to_text('insert',$$active_text_ref,encode_base64($pdata));
    $main::dreaminfo{'entry-changed'}=1;
    $main::dreaminfo{'db-dirty'}=1;
  }

sub add_image_from_web
  {
    my ($active_text_ref)=@_;
    my $f1;
    my $message='Enter URL of image';
    my $url;
    my $idata;
    while(1)
      {
	my $url_dialog = $main::toplevel -> 
	  DialogBox( -bg => $main::dream_options{'bgcolor'},
		     -fg => $main::dream_options{'fgcolor'},
		     -title => $message,
		     -buttons => ['OK','Cancel']
		   );
	
	$url_dialog -> add('Entry',
			   -width => 80,
			   -bg =>$main::dream_options{'quesbg'},
			   -fg => $main::dream_options{'quesfg'},
			   -textvariable => 
			   \$url)->pack();
	my $response = $url_dialog-> Show();
	return if($response eq 'Cancel');
	$idata = LWP::Simple::get($url);
	last if($idata);
	$message = 'Unable to read this URL, re-enter';
      }
    my $pdata;
    unless($main::windows_os)
      {
	my $imag = Image::Magick->new();
	my $blobs = $imag -> BlobToImage(($idata));
	
	$pdata=($imag->ImageToBlob(magick=>'JPEG'))[0];
	return unless($pdata);
      }
    else
      {
	$pdata=$idata;
      }
    post_image_to_text('insert',$$active_text_ref,encode_base64($pdata));
    $main::dreaminfo{'entry-changed'}=1;
    $main::dreaminfo{'db-dirty'}=1;
  }



sub color_scheme 
  {
  my $mainwindow= $main::toplevel
    ->DialogBox(
		-title=> 'Press the buttons below to choose colors',
	       );

    my(@pl) = qw/-side top -anchor c -pady 2m/;
    my($back, $front,$dbg,$ibg,$itext);

    $back = $mainwindow->add('Button',-text => 'Set background color',
				-bg =>$main::dream_options{'bgcolor'},
				-fg => $main::dream_options{'fgcolor'})
      ->pack(@pl);

    $back->configure(-command => 
		     [\&setColor,$back,'bgcolor']
		      );
  
  $front = $mainwindow->add('Button',
			    -text => 
			    'Set foreground color',
			    -bg =>$main::dream_options{'bgcolor'},
			    -fg => $main::dream_options{'fgcolor'})
    ->pack(@pl);
 
  $front->configure(-command => [\&setColor,$front,'fgcolor',]
		      );
  $dbg = $mainwindow->add('Button',
			    -text => 
			    'Set dream background color',
			    -bg =>$main::dream_options{'bgcolor'},
			    -fg => $main::dream_options{'fgcolor'})
    ->pack(@pl);
 
  $dbg->configure(-command => [\&setColor, $dbg,'dreambg',]
		      );

  $ibg = $mainwindow->add('Button',
			    -text => 
			    'Set interpretation background color',
			    -bg =>$main::dream_options{'bgcolor'},
			    -fg => $main::dream_options{'fgcolor'})
    ->pack(@pl);
 
  $ibg->configure(-command => [\&setColor,$ibg,'interpbg',]
		      );

  $itext = $mainwindow->add('Button',
			    -text => 
			    'Set interpretation text color',
			    -bg =>$main::dream_options{'bgcolor'},
			    -fg => $main::dream_options{'fgcolor'})
    ->pack(@pl);
 
  $itext->configure(-command => [\&setColor,$itext, 'interpgf',]
		      );



  $mainwindow ->Show();
  post_colors();

  my $response = communicate('Keep these colors?',
			['Yes','No','Cancel']);

  if($response eq 'Yes')
    {
      save_options();
    }
  else
    {
      default_color_scheme();
    }

}


sub setColor {
  my($button, $option) = @_;
  my $color;
  if($main::windows_os)
    {
      $color = $button->chooseColor();
    }
  else
    {
      $color = $button->chooseColor(
				    -bg =>$main::dream_options{'bgcolor'},
				    -fg => $main::dream_options{'fgcolor'});
    }
  return unless(defined $color);
  $main::dream_options{$option}=$color;  
}


sub new_passphrase
  {
    my($passphrase_ref)=@_;
    my $passphrase1;
    my $passphrase2;
    my $message = 'Enter a new pass-phrase:';

    while(1)
      {
	my $passphrase_dialog = $main::toplevel -> 
	  DialogBox( -bg => $main::dream_options{'bgcolor'},
		     -fg => $main::dream_options{'fgcolor'},
		     -title => "Enter passphrase",
		     -buttons => ['OK','Cancel']
		   );
	$passphrase_dialog -> add('Label',
				  -width => 60,
				  -bg =>$main::dream_options{'bgcolor'},
				  -fg => $main::dream_options{'fgcolor'},
				  -text=>
				  $message)->pack();
	$passphrase_dialog -> add('Entry',
				  -width => 60,
				  -bg =>$main::dream_options{'quesbg'},
				  -fg => $main::dream_options{'quesfg'},
				  -textvariable => 
				  \$passphrase1,
				  -show => '*')->pack();
	$passphrase_dialog -> add('Label',
				  -width => 60,
				  -bg =>$main::dream_options{'bgcolor'},
				  -fg => $main::dream_options{'fgcolor'},
				  -text=>
				  'Enter passphrase a second time')->pack();
	$passphrase_dialog -> add('Entry',
				  -width => 60,
				  -bg =>$main::dream_options{'quesbg'},
				  -fg => $main::dream_options{'quesfg'},
				  -textvariable => 
				  \$passphrase2,
				  -show=>'*')->pack();
	my $result = $passphrase_dialog->Show();
	if($result eq 'Cancel')
	  {
	    exit() unless($$passphrase_ref =~ m/\S/);
	    return '';
	  };
	$passphrase1 =~ s/\s+/\ /sg;
	$passphrase1 =~ s/^\s*(\S.*\S)\s*\Z/$1/sg;
	$passphrase2 =~ s/\s+/\ /sg;
	$passphrase2 =~ s/^\s*(\S.*\S)\s*\Z/$1/sg;
	last if($passphrase1 eq $passphrase2);
	$message = 'Your entries did not match. Please re-enter';
      }
    $$passphrase_ref=$passphrase1;
    return $passphrase1;
  }

sub decrypt_file
  {
    my ($db_in)=@_;
    my $passphrase='';
    my $message = 'Please enter your pass-phrase';
    while(1)
      {
	my $passphrase_dialog = $main::toplevel -> 
	  DialogBox( -bg => $main::dream_options{'bgcolor'},
		     -fg => $main::dream_options{'fgcolor'},
		     -title => $message,
		     -buttons => ['OK','Cancel']
		   );
	my $pass_entry=$passphrase_dialog -> add('Entry',
				  -width => 60,
				  -bg =>$main::dream_options{'quesbg'},
				  -fg => $main::dream_options{'quesfg'},
				  -textvariable => 
				  \$passphrase,
				  -show=>'*')->pack();
	$passphrase_dialog->configure(-focus=> $pass_entry);
	my $result = $passphrase_dialog->Show();
	return '' if($result eq 'Cancel');

	$passphrase =~ s/\s+/\ /sg;
	$passphrase =~ s/^\s*(\S.*\S)\s*\Z/$1/sg;
	
	my $cipher = Crypt::CBC->new(
				     -key => $passphrase,
				     -cipher => 'Blowfish',
				     -padding=> 'space',
				     -salt=>1,
				     -header  => 'salt',
				    )
	  || die "Couldn't create CBC object"; 
	
	my $compressed_string = $cipher->decrypt($db_in);
	my $plain_string;
	gunzip \$compressed_string => \$plain_string
	  or die "Unable to uncompress database\n";
	last if(eval($plain_string));
	$message = 'Your entry was invalid. Please re-enter';
      }
    return $passphrase;
  }

sub call_spelling
{
    my (
	$dream_text,
	$dream_interp,
	) = @_;
    spelling($dream_text);
    spelling($dream_interp);
    $main::dreaminfo{'entry-changed'}=1;
    $main::dreaminfo{'db-dirty'}=1;
    post_pending_changes();
}


sub spelling
{
    my (
	$text_handle,
	) = @_;
    my $speller = Text::Aspell->new;
    my $retval=0;
    my $last_index = $text_handle -> index('end');
    my ($nlines, $nchars_last_line) = split /\./, $last_index;
    my $lineno;
    my $ispell_results;
    my $action = '';
    LINE:for ($lineno = 1; $lineno <= $nlines; $lineno++)
    {
	my $lineoftext = $text_handle -> get("$lineno.0","$lineno.end");



	my $replace_word = '';
	while($lineoftext =~ m/(\w+)/g)
	{
	  my $word=$1;
	  next if($speller->check($word));
	  $retval=1;
	  my $offset = pos($lineoftext) - length($word);
	  my @words = $speller -> suggest($word);
	  my $count = @words;
	  $text_handle -> 
	    tagAdd('sel',"$lineno.$offset","$lineno.$offset wordend");
	  spell_dialog(
		       $word,
		       \@words,
		       $count,
		       \$replace_word,
		       \$action,
		       $speller);
	  $text_handle -> 
	    tagRemove('sel',"$lineno.$offset",
		      "$lineno.$offset wordend");
	  if ($action eq 'Replace')
	    {
	      if ($word =~ m/\'/)
		{
		  $text_handle ->
		    delete("$lineno.$offset",
			   "$lineno.$offset wordend + 2 chars");
		}
		else 
		{
		    $text_handle ->
			delete("$lineno.$offset",
			       "$lineno.$offset wordend");
		}
	
		$text_handle -> 
		    insert("$lineno.$offset",$replace_word)
			if ($replace_word =~ /\S/);
		redo LINE;
	    }
	}
    }
    $speller->save_all_word_lists;
    return $retval;
  }


sub spell_dialog
{
   my
	(
	 $misspelled,
	 $words_ref,
	 $count,
	 $replace_word_ref,
	 $action_ref,
	 $speller
	 ) =@_;
    my $done = 0;
    my $sdia = $main::toplevel -> 
      DialogBox(-title => "Correct word",
		-buttons => ["Ignore","Replace",
			     "Add to dictionary"],
		-bg => $main::dream_options{'bgcolor'},
		-fg => $main::dream_options{'fgcolor'});
   $sdia -> Label(-text => $misspelled,
		   -bg => $main::dream_options{'bgcolor'},
		   -fg => $main::dream_options{'fgcolor'}) -> pack();
    my $replace_entry = $sdia -> Entry(
				       -bg => $main::dream_options{'quesbg'},
				       -fg => $main::dream_options{'quesfg'},)
	-> pack();
    my $replace_list = $sdia -> Listbox(-bg => $main::dream_options{'quesbg'},
				       -fg => $main::dream_options{'quesfg'},)
	-> pack();
    foreach my $word(@{$words_ref})
    {
	$replace_list -> insert('end',$word);
    }
    $replace_list -> bind('<Double-1>',sub{
	$replace_entry -> delete(0,'end');
	my $rword = $replace_list -> get('active');
	$replace_entry -> insert(0,$rword);
    });
    $$action_ref = $sdia -> Show();
    return if($$action_ref eq 'Ignore');
    if ($$action_ref eq 'Replace')
    {
	$$replace_word_ref = $replace_entry -> get();
	$speller -> store_replacement($misspelled, $$replace_word_ref);
	return;
    }
   $speller->add_to_session($misspelled);
   $speller->add_to_personal($misspelled);
}


sub style_text
  {
    my (
	$in_widget,
	$style,
       ) =@_;
    my $widget=$in_widget;
    $widget=$$in_widget if(ref($in_widget) eq 'REF');
    return unless (ref($widget) eq 'Tk::TextUndo');
    my @style_already =
      $widget -> tagNextrange($style, 'sel.first','sel.last');


######################special validation for active dates
    if($style eq 'active_date') # must validate the date
      {
	my $new_date;
	my $dtext = $widget -> get('sel.first','sel.last');

	unless($new_date=UnixDate(ParseDate($dtext),'%Y%m%d'))
	  {
	    my $date_message=
	      $main::toplevel->Dialog( 
			      -bg => $main::dream_options{'bgcolor'},
			      -fg => $main::dream_options{'fgcolor'},
			      -text => 'Not a valid date',
			      -buttons => ['OK','Cancel']);
	    my $result = $date_message->Show();
	    return;
	  };
	unless (exists $main::dreams_bydate{$new_date})
	  { 
	    my $date_message=
	      $main::toplevel->Dialog( 
			      -bg => $main::dream_options{'bgcolor'},
			      -fg => $main::dream_options{'fgcolor'},
			      -text => 'No entry at that date',
			      -buttons => ['OK','Cancel']);
	    my $result = $date_message->Show();
	    return;
	  };
	
      }
    $widget -> tagAdd($style,'sel.first','sel.last');
    $widget ->
      tagRemove($style,$style_already[0],$style_already[1])
	if (@style_already > 0);
    $main::dreaminfo{'entry-changed'} = 1;
    $main::dreaminfo{'db-dirty'}=1;
    
  }


sub rescan_button
{
    my (
	$parent_ref,
	) = @_;
    require Tk::DialogBox;
    my $search_words;
    my $search_dialog = $parent_ref -> 
      DialogBox( -bg => $main::dream_options{'bgcolor'},
		 -fg => $main::dream_options{'fgcolor'},
		 -title => "Enter Search Words",
		 -buttons => ['OK','Cancel']
					   );
    my $sentry=$search_dialog -> add('Entry',
			-width => 50,
			-bg =>$main::dream_options{'quesbg'},
			-fg => $main::dream_options{'quesfg'},
			-textvariable => 
			\$search_words)->pack();
    $search_dialog->configure(-focus=>$sentry);
    my $result = $search_dialog->Show();
    return if($result eq 'Cancel');

    search_word_set($search_words);
}

sub search_exact_phrase_button
{
    my (
	$parent_ref,
	) = @_;
    require Tk::DialogBox;
    my $search_words;
    my $search_dialog = $parent_ref -> 
      DialogBox( -bg => $main::dream_options{'bgcolor'},
		 -fg => $main::dream_options{'fgcolor'},
		 -title => "Enter phrase",
		 -buttons => ['OK','Cancel']
					   );
    my $sentry=$search_dialog -> add('Entry',
			-width => 50,
			-bg =>$main::dream_options{'quesbg'},
			-fg => $main::dream_options{'quesfg'},
			-textvariable => 
			\$search_words)->pack();
    $search_dialog->configure(-focus=>$sentry);
    my $result = $search_dialog->Show();
    return if($result eq 'Cancel');

    search_exact_phrase($search_words);
}


sub active_date
{
    my (
	$text_handle,  #implicit parameter
	) = @_;

   
    my ($dbegin,$dend) = $text_handle ->
	tagPrevrange('active_date','current');
    return unless($dbegin and $dend);
    my $search_date_info = $text_handle ->
	get($dbegin,$dend);
    return unless($search_date_info);

# don't do anything unless the selection is OK


    $main::back_button -> configure(
				    -text=>'Go Back',
				    -state=>'normal',
				    -image => $main::back_image);

    # get out of any pending searches
    my $stack_size = @main::date_stack;
    while ($stack_size > 1)
      {
	pop @main::date_stack;
	$stack_size = @main::date_stack;
      }
    my $dates_inorder = $main::date_stack[-1]->{'range'};
# duplicate the top of the stack
    push @main::date_stack,({'range'=>$dates_inorder,
			     'place'=>0});

    my $search_date = UnixDate(
			       ParseDate($search_date_info),'%Y%m%d');

    access_dream($search_date);
}

sub range_word_set
  {
    my ($word_list)=@_;
    my @keywords = split /\W/, $word_list;
    my @dateset =  @{$main::date_stack[-1]->{'range'}};
    my %set1;
    foreach my $date (@dateset)
      {
	$set1{$date}=1;
      }
    
    foreach my $word (@keywords)
      {
	my $lc_word = clean_word($word);
	next if(short_word($lc_word));
	foreach my $sdate (@dateset)
	  {
	    unless(exists $main::word_date{$lc_word}->{$sdate})
	      {
		delete $set1{$sdate};
	      };
	  }
	@dateset = sort keys %set1;
      }
    return \@dateset;
  }


sub search_exact_phrase
  {
    my (
	$word_list)=@_;
    my $dateset_ref = range_word_set($word_list);
    my $found_count = @{$dateset_ref};
    unless($found_count)
      {
	communicate("Phrase: \'$word_list\' not found");
	return;
      };
# convert into a regular expression
    $word_list =~ s/^\s*(.*)\s*$/$1/gs;
    my $reg_word_list = $word_list;
    $reg_word_list =~ s/\s+/\\s\+/gs;
    my @new_date_list=();

    foreach my $date (@{$dateset_ref})
      {
	my $raw_text = $main::dreams_bydate{$date} ->[0];
	#remove all tags
	$raw_text =~ s/(mark|current|text|\d+\D\d+|anchor)<tag>/ /gis;
	push @new_date_list,($date) 
	  if($raw_text =~ m/$reg_word_list/gis);
      }
    $found_count = @new_date_list;
    unless($found_count)
      {
	communicate("Phrase: \'$word_list\' not found");
	return;
      };

    # now we have a set of dates including the phrase
    # Narrow the focus to those dates only.
    search_narrow_focus(\@new_date_list);
  }

sub search_narrow_focus
  {
    my ($date_list_ref)=@_;
    $main::back_button -> configure(
				    -text=>'Undo Search',
				    -state=>'normal',
				    -image=>$main::up_image);
    push @main::date_stack, ({'range'=>$date_list_ref,
			      'place'=>0});
    my $found_count = @{$date_list_ref};
    $main::dscale -> configure(
			      -to => $found_count-1,
			      -tickinterval => set_ticks($found_count),
			      -variable => \$main::date_stack[-1]->{'place'},
			     );
    post_pending_changes(); 
  }

sub search_word_set
{
    my ($word_list)=@_;
    my $dateset_ref = range_word_set($word_list);
    my $found_count = @{$dateset_ref};
    unless($found_count)
      {
	communicate("Words: \'$word_list\' not found");
	return;
      };
# now we have a set of dates common to all words
# Narrow the focus to those dates only.
    search_narrow_focus($dateset_ref);
  }

sub active_quote
{
    my (
	$text_handle,  #implicit parameter
	) = @_;
    my ($dbegin,$dend) = $text_handle ->
	tagPrevrange('quotecolor','current');
    my $search_words = $text_handle ->
	get($dbegin,$dend);

    search_word_set($search_words);
}

sub undo_search
  {
    my $size = @main::date_stack;
    if($size <= 2)
      {
	$main::back_button -> configure(-state => 'disabled',
					-text =>'',
					-image=>'');
      }
    return if ($size < 2);
    pop @main::date_stack;
    my $ndreams = @{$main::date_stack[-1]->{'range'}};
    $main::dscale -> configure(
			  -to => $ndreams-1,
			  -tickinterval => set_ticks($ndreams),
			  -variable => \$main::date_stack[-1]->{'place'},
			 );
    post_pending_changes();
  }

sub get_date_goto_dream
  {
    my (
	undef,
       )=@_;

# post any pending changes
    post_pending_changes();
    my $entered_date;
    my $message = 'Enter the date you want to access:';
    my $new_date;

    while(1)
      {
	require Tk::DialogBox;
	my $date_dialog = $main::toplevel -> 
	  DialogBox( -bg => $main::dream_options{'bgcolor'},
		     -fg => $main::dream_options{'fgcolor'},
		     -title => "Enter Date",
		     -buttons => ['OK','Cancel']
		   );
	$date_dialog -> add('Label',-text=>$message,
			    -bg => $main::dream_options{'bgcolor'},
			    -fg => $main::dream_options{'fgcolor'},
			   )->pack();
	my $entry_widget=$date_dialog -> add('Entry',
			    '-bg' =>$main::dream_options{'quesbg'},
			    '-fg' => $main::dream_options{'quesfg'},
			    -textvariable => 
			    \$entered_date)->pack();
	$date_dialog -> configure(-focus=>$entry_widget);
	my $result = $date_dialog->Show();
	return if($result eq 'Cancel');
	$new_date=UnixDate(ParseDate($entered_date),'%Y%m%d');
	last if($new_date);
	$message = 'Your entry was invalid, please re-enter:';
      }
    access_dream($new_date);
  }

sub go_to_date
  {
    my ($dream_date)=@_;
    # this assumes a dream already exists at this date

    my $dream_number = 0;
    my $dates_inorder_ref = $main::date_stack[-1]->{'range'};
    foreach my $date (@{$dates_inorder_ref})
      {
	last if($date eq $dream_date);
	$dream_number++;
      }
    $main::date_stack[-1]->{'place'}=$dream_number;
  }

sub initialize_new_dream
  {
    my (
	$dream_date
       ) = @_;

# no dream can exist at this date!
    my $new_dream = [
		     '',   #dream
		     '',   #interp
		     5];   #intensity
    my $dream_number;
    $main::dreams_bydate{$dream_date}=$new_dream;
# make local copy
    my $dates_inorder_ref = $main::date_stack[-1]->{'range'};
    

    # locate the new date in the sorted list
    if($dream_date lt $dates_inorder_ref->[-1])
      {
	# find the new dream's position in the list
	$dream_number=0;
	foreach my $sdate (@{$dates_inorder_ref})
	  {
	    last if($sdate gt $dream_date);
	    $dream_number++;
	  }
	# insert the new dream date into the sorted array
	splice(@{$dates_inorder_ref},$dream_number,0,$dream_date);
      }
    else
      {
	push @{$dates_inorder_ref}, ($dream_date);
	$dream_number=@{$dates_inorder_ref}-1;
      }
# the number of dreams has definitely changed...
    my $ndreams =@{$dates_inorder_ref};
	
    $main::dscale -> configure(
			      -to => $ndreams-1,
			      -tickinterval => set_ticks($ndreams),
			     );
# post modified local copy to the global stack
    $main::date_stack[-1]->{'range'}= $dates_inorder_ref;
    $main::date_stack[-1]->{'place'}= $dream_number;
  }

sub access_dream_key2
  {
    my (
	undef,
	$dream_date_ref) =@_;
    access_dream($$dream_date_ref);
  }

sub access_dream_key
  {
    my (
	undef,
	$dream_date
       ) = @_;
    # get out of any pending searches
    my $stack_size = @main::date_stack;
    while ($stack_size > 1)
      {
	pop @main::date_stack;
	$stack_size = @main::date_stack;
      }
# in case the date stack has changed...
    my $ndreams = @{$main::date_stack[-1]->{'range'}};

    $main::dscale -> configure(
			      -to => $ndreams-1,
			      -tickinterval => set_ticks($ndreams),
			      -variable => \$main::date_stack[-1]->{'place'},
			     );
    access_dream($dream_date);
  }

sub access_dream
{
    my (
	$dream_date
       ) = @_;

    if(exists $main::dreams_bydate{$dream_date})
      {
	go_to_date($dream_date);
	# the dream date isn't really new...
      }
    else
      {
	initialize_new_dream($dream_date);
      }

    post_pending_changes();
}

sub post_pending_changes
{
    my $dates_inorder_ref = $main::date_stack[-1]->{'range'};
    my $ndreams = @{$dates_inorder_ref};
    my $old_date = $main::dreaminfo{'date'};
    my $new_date_info;
    if($main::DEBUG)
      {
	print "entered post_pending_changes\n";
	print "ndreams=$ndreams,\n";
	print "old_date=$old_date,\n"; 
	print "main::dreaminfo{\'entry_changed\'}=",
	  $main::dreaminfo{'entry-changed'},",\n";
      }

    
    if ($main::dreaminfo{'entry-changed'})
      {
	update_text($old_date);
	update_interp($old_date);
	update_intensity($old_date);
	$main::dreaminfo{'entry-changed'}=0;
	$main::dreaminfo{'db-dirty'}=1;
      }

# now display information for the current entry
    my $dream_number = $main::date_stack[-1]->{'place'};
    my $newdate = $dates_inorder_ref->[$dream_number];
    $main::dreaminfo{'date'}=$newdate;
    if($main::DEBUG)
      {
	print "dream_number=$dream_number,\n";
	print "newdate=$newdate,\n"; 
	print "---post_pending_changes\n";
      }

    $main::dream_numeric = $main::dreams_bydate{$newdate}
      ->[2];
    my $desc = $main::dream_description_array[$main::dream_numeric];
    $main::dream_intensity
      ->configure(
		  -text=>$desc,
		  -fg => $main::dream_options{'dreamfg'} -> [$main::dream_numeric],
		  -bg => $main::dream_options{'dreambg'} -> [$main::dream_numeric]
		 );
   
    $main::dream_text -> delete('1.0','end');
    $main::dream_interp -> delete('1.0','end');
    $main::dreaminfo{'number_ref'} = \$dream_number;
    $main::dreaminfo{'date'} = $newdate;
    my $dr_text = $main::dreams_bydate{$newdate}
      ->[0];
    post_encoded($dr_text,$main::dream_text);
    my $dr_interp =$main::dreams_bydate{$newdate}
      ->[1];
    post_encoded($dr_interp,$main::dream_interp);
    my $pretty_date = UnixDate($newdate,'%A, %b %e %Y');
    die "Date error! Date=$pretty_date|\n" unless($pretty_date);
    $main::entered_date = $pretty_date;
    $main::dream_text ->
      configure(-foreground => 
		$main::dream_options{'dreamfg'}->[$main::dream_numeric],
		-background => 
		$main::dream_options{'dreambg'}->[$main::dream_numeric]
	       );
  }


sub delete_dream
  {
    my(
       $dreams_bydate_ref,
       )=@_;
    my $response = communicate("Want to delete this entry?");
    return if($response eq 'Cancel');
    my $dates_inorder_ref = $main::date_stack[-1]->{'range'};
    my $dream_number = $main::date_stack[-1]->{'place'};
    my $date = $dates_inorder_ref ->[$dream_number];
    my $dream_text = $dreams_bydate_ref->{$date}->[0];
    delete_coded_text_from_word_array($dream_text,$date);
    delete $dreams_bydate_ref->{$date};
    create_dates($dates_inorder_ref);
    $main::date_stack[-1]->{'range'}=$dates_inorder_ref;
    my $ndreams = @{$dates_inorder_ref};
    $dream_number-- if($dream_number > $ndreams - 1);
    $main::dscale -> configure(
			      -to => $ndreams-1,
			      -tickinterval => set_ticks($ndreams));
    $main::dreaminfo{'db-dirty'}=1;
    $main::date_stack[-1]->{'place'}=$dream_number;
    post_pending_changes();
  }

sub delete_coded_text_from_word_array
  {
    my ($dream_text,$date)=@_;
    my @text_array = split /\<tag\>/,$dream_text;
    my $tagsize = @text_array;
    if ($tagsize == 1)
      {
	delete_words_from_date(decode_string($dream_text),$date);
      }
    else
      {
	my $i;
	for ($i =0; $i < $tagsize; $i += 3)
	  {
	    if(decode_string($text_array[$i]) =~ m/text/)
	      {
		delete_words_from_date(
				       decode_string(
						     $text_array[$i+1]),
				       $date);
	      }
	  }
      }
  }

sub update_text
  {
    my ($date)=@_;
    my $old_text = $main::dreams_bydate{$date}->[0];
    delete_coded_text_from_word_array($old_text,$date);
    my @tagdump =  $main::dream_text
      -> dump('0.0','end');
    my $new_text='';
    my $prev_key='';

    foreach my $tstring (@tagdump)
      {
	post_words_to_date($tstring,
			   $date)
	  if($prev_key =~ m/text/);
	my $encoded_range = encode_string($tstring);
	$encoded_range = encode_string('image')
	  if($tstring eq 'window');
	if($prev_key eq 'image')
	  {
	    my $image_data = $tstring->data(-format=>'jpeg');
	    $encoded_range = encode_string($image_data);
	  }
	if($prev_key eq 'window')
	  {
	    my $button_callback = $tstring->cget(-command);
	    my $image_data = $button_callback -> [1]; 
	    $encoded_range = encode_string($image_data);
	  }
	$new_text .= "$encoded_range<tag>";
	$prev_key = $tstring;
      }
#trim space at the bottom
    $main::dreams_bydate{$date}->[0]=trim_tagged($new_text);
  }


sub update_interp
  {
    my ($date)=@_;
    my @tagdump =  $main::dream_interp
      -> dump('0.0','end');
    my @coded_tagdump = ();
    my $tagsize = @tagdump;
    my $prev_key='';
    my $new_interp;
    foreach my $tstring (@tagdump)
      {
	my $encoded_range = encode_string($tstring);
	$encoded_range = encode_string('image')
	  if($tstring eq 'window');
	if($prev_key eq 'image')
	  {
	    my $image_data = $tstring->data(-format=>'jpeg');
	    $encoded_range = encode_string($image_data);
	  }
	if($prev_key eq 'window')
	  {
	    my $button_callback = $tstring->cget(-command);
	    my $image_data = $button_callback -> [1]; 
	    $encoded_range = encode_string($image_data);
	  }
	$new_interp .= "$encoded_range<tag>";
	$prev_key = $tstring;
      }
# trim space at the bottom
    $main::dreams_bydate{$date}->[1]=trim_tagged($new_interp);
  }

sub trim_tagged
  {
    my ($in_text)=@_;
    $in_text  =~ s/(<tag>\d+\D\d+<tag>\s*|<tag>\d+\D\d+<tag>text<tag>\s*|<tag>\d+\D\d+<tag>mark<tag>(insert|anchor|notepos))*$//gs;
    $in_text  =~ s/\s*$//gs;
    return $in_text;
  }
    
sub update_intensity
  {
    my ($date)=@_;
    $main::dreams_bydate{$date}->[2]=$main::dream_numeric;
  }
  
sub post_words_to_date
  {
    my ($in_str,$date)=@_;
    my @inwords = split /\W+/, $in_str;
    foreach my $word (@inwords)
      {
	my $lc_word = clean_word($word);
	next if short_word($lc_word);
	if(exists $main::word_date{$lc_word})
	   {
	     my $date_set = $main::word_date{$lc_word};
	     $date_set -> {$date}=1
	       unless(exists $date_set -> {$date})
	   }
	else
	  {
	    $main::word_date{$lc_word}={$date => 1};
	  }
      }
  }
sub short_word
  {
    my ($in_word)=@_;
    return 1 unless($in_word =~ m/\S/);
    return 1 if ($in_word =~ m/^\s*\w\s*\Z/);
    return 0 if(length($in_word)>3);
    return 1 if($in_word =~ m/^\s*(the|in|on|an|that|they)\s*\Z/);
    return 0;
  }

sub clean_word
  {
    my ($in_word)=@_;
    my $first = lc($in_word);
    $first =~ s/[\.\'\"\x{2019}\x{201c}\x{201d}\x{2014}\,\)\(\:]+//g;
    return $first;
  }


sub create_dates
  {
    my ($dates_inorder_ref)=@_;
    @{$dates_inorder_ref}=
      sort  keys %main::dreams_bydate;
    unless(ParseDate($dates_inorder_ref->[0]))
      {
	shift @{$dates_inorder_ref};
	delete $main::dreams_bydate{''};
      }
  }


sub post_encoded
{
    my (
	$coded_text,
	$text_widget
	) = @_;
    return unless($coded_text);

    my @text_array = split /\<tag\>/,$coded_text;
    my $tagsize = @text_array;
    if ($tagsize == 1) # there are no tags
    {
      $text_widget ->
	insert('1.0',trim_string(decode_string($coded_text)));
    }
    else
    {
	my $i;
        my @current_tags = ();
      TAGPROC:for ($i =0; $i < $tagsize; $i += 3)
      {
	  my $new_coded_text = $text_array[$i];
	  if ($new_coded_text =~ m/tagon/)
	  {
	      my $tag_type = $text_array[$i+1];
	      push @current_tags,("$tag_type")
		  if ($tag_type ne 'sel');
	      next TAGPROC;
	  }
	  if ($new_coded_text =~ m/tagoff/)
	  {
	      pop @current_tags if ($text_array[$i+1] ne 'sel');
	      next TAGPROC;
	  }
	  if($new_coded_text =~ m/text/)
	    {
	      $text_widget ->
		insert('end',decode_string($text_array[$i+1]),
		       @current_tags);
	    }
	  if($new_coded_text =~ m/image/)
	    {
	      my $image_data = decode_string($text_array[$i+1]);
	      post_image_to_text(
				 'end',
				 $text_widget,
				 $image_data,
				);
	    }
	}
      }
  }

sub post_image_to_text
  {
    my($where,
       $text_widget,
       $image_data)=@_;
    my $new_image = $text_widget -> Photo(-data=>$image_data);
    my $scalesize = $text_widget -> pixels('60m');
    unless($new_image)
      {
       print "bad image. data=$image_data\n";
       return;
      };
    if($new_image->height>$scalesize)
      {
	post_button_to_text($where,
			    $text_widget,
			    $image_data,
			    $new_image,
			    $scalesize);
      }
    else
      {
	$text_widget -> imageCreate($where,-image=>$new_image,
				    -padx=>10,-pady=>10,
				    -align=>'center');
      }
  }

sub post_button_to_text
  {
    my ($where,
	$text_widget,
	$image_data,
	$new_image,
	$scalesize)=@_;
    my $new_button;
    my $aspect_ratio = ($new_image -> width)/($new_image->height);
    my $shrink_image = $main::toplevel 
      ->Photo();
    $shrink_image -> copy($new_image,-subsample=>
			  int($new_image->height/$scalesize+.5));
    
    my $new_button_com = qq!\$new_button = \$text_widget 
      -> Button(-image=>\$shrink_image,
                -relief => 'raised',
                -borderwidth => 5,
                -background=>'blue',
		-command =>[
			    sub{
                              my \$idata = shift; 
			      my \$wind = \$main::toplevel
				->Toplevel();
			      
			      my \$d_image = \$wind
				-> Photo(-data=>\$idata);
			      my \$l = \$wind->Label(-image =>\$d_image)
                                -> pack();
			    },'! .$new_image->data(-format=>'JPEG')."\']);";
    eval($new_button_com);
    
   $text_widget -> window('create',$where,-window=>$new_button);
}

sub delete_words_from_date
  {
    my ($tstring,$ddate)=@_;
    my @inwords = split /\s+/, $tstring;
    foreach my $word (@inwords)
      {
	my $lc_word = clean_word($word);
	next if (short_word($lc_word));
	if(exists $main::word_date{$lc_word})
	   {
	     my $date_set = $main::word_date{$lc_word};
	     delete $date_set -> {$ddate}
	       if(exists $date_set -> {$ddate});
	   }
      }
  }


sub decode_string
  {
    my ($in_text)=@_;
    $in_text =~ s/\&\#([^\;]+)\;/tounicode($1)/sge;
    $in_text =~ s/\&mdash;/\x{2014}/sgo;
    $in_text =~ s/\&ndash;/\x{2013}/sgo;
    $in_text =~ s/\&lsquo;/\x{2018}/sgo;
    $in_text =~ s/\&rsquo;/\x{2019}/sgo;
    $in_text =~ s/\&ldquo;/\x{201c}/sgo;
    $in_text =~ s/\&rdquo;/\x{201d}/sgo;
    $in_text =~ s/\&quot;/\'/sgo;

#unescape <>
    $in_text =~ s/\&gt;/>/sgo;
    $in_text =~ s/\&lt;/</sgo;

# unescape &
    $in_text =~ s/\&amp;/\&/sgo;
    return $in_text;
  }

sub tounicode
  {
    my ($hex_in)=@_;
    return pack("U*",hex($hex_in));
  }



sub encode_string
  {
    my ($in_text)=@_;
    $in_text =~ s/\&/\&amp;/sgo;
    $in_text =~ s/\x{201d}/\&rdquo;/sgo;
    $in_text =~ s/\x{201c}/\&ldquo;/sgo;
    $in_text =~ s/\x{2019}/\&rsquo;/sgo;
    $in_text =~ s/\x{2018}/\&lsquo;/sgo;
    $in_text =~ s/\x{2014}/\&mdash;/sgo;
    $in_text =~ s/\x{2013}/\&ndash;/sgo;
    $in_text =~ s/---\s/\&mdash; /sgo;
    $in_text =~ s/--\s/\&ndash; /sgo;
    $in_text =~ s/>/\&gt;/sgo;
    $in_text =~ s/</\&lt;/sgo;

# smart quotes
    $in_text =~ s/\"(\w)/\&ldquo\;$1/sgo;
    $in_text =~ s/\"/\&rdquo\;/sgo;
    $in_text =~ s/(\w)\'(\w)/$1\&rsquo\;$2/sgo;
    $in_text =~ s/\'(\w)/\&lsquo\;$1/sgo;
    $in_text =~ s/(\w)\'/$1\&rsquo\;/sgo;

    $in_text =~ s/\'/\&quot;/sgo;

# handle any other unicode chars
    return join("",
		       map { $_ > 128 ?                  # if wide character...
			       sprintf("&\#x%x;", $_) :  # \x{...}
				 chr($_)                     # else as themselves
			       } unpack("U*", $in_text)
		      );           # unpack Unicode characters

    return $in_text;
}

sub trim_string
  {
    my ($in_string)=@_;
    $in_string =~ s/\s*(\S.*)/$1/gs;
    $in_string =~ s/(.*\S)\s*/$1/sg;
    return $in_string;
  }


sub save_button
{
    my
	(undef,
	 ) = @_;
    $main::toplevel -> Busy(-recurse=>1);
  #update everything
    post_pending_changes();
    my $cipher = Crypt::CBC->new(
				 -key => $main::dreaminfo{'passphrase'},
				 -cipher => 'Blowfish',
				 -padding=> 'space',
				 -salt =>1,
				 -header  => 'salt',
				)
      || die "Couldn't create CBC object"; 

    my $compressed_string;
    my $complete_string = 
      Data::Dumper->Dump([\%main::dreams_bydate], 
			 ["*main::dreams_bydate"]) . ";\n" .
	Data::Dumper->Dump([\%main::word_date],["*main::word_date"]);
# print $complete_string;
    gzip \$complete_string => \$compressed_string
      or die "gzip failed: $GzipError\n";
    my $db_filename=$main::dreaminfo{'file-name'};
    unless($main::dreaminfo{'file-name'} =~ m/\S/)
      {
	if($main::windows_os)
	  {
	    $db_filename = $main::toplevel 
	      -> getSaveFile(-initialfile=>'dreamdb',
			    );
	  }
	else
	  {
	    $db_filename = $main::toplevel 
	      -> getSaveFile(-initialfile=>'dreamdb',
			     -bg => $main::dream_options{'interpbg'},
			     -fg => $main::dream_options{'interpfg'},
			    );
	  }
      }
	my $file_found = 0;
    foreach my $entry (@{$main::dream_options{'file-history'}})
      {
	$file_found=1 if($db_filename eq $entry);
      }
    unless($file_found)
      {
	push @{$main::dream_options{'file-history'}},($db_filename);
	shift @{$main::dream_options{'file-history'}} 
	  if(@{$main::dream_options{'file-history'}}>4);
	save_options();
      }
    unlink $main::tempfilename if($main::tempfilename);
    ($main::tempfilehand,$main::tempfilename)=tempfile();
    binmode $main::tempfilehand;;
    print $main::tempfilehand $cipher->encrypt($compressed_string);
    close $main::tempfilehand;
    
    $main::dreaminfo{'file-name'}=$db_filename;
    $main::toplevel -> Unbusy();
    $main::dreaminfo{'db-dirty'}=0;
    $main::tmpwritten=1;
  }

sub save_options
  {
    open CFILE,">$main::options_file";
    print CFILE Data::Dumper->Dump([\%main::dream_options],
				   ['*main::dream_options']);
    close CFILE;
  }

sub set_ticks
  {
    my ($ndreams)=@_;
    return 1 if ($ndreams < 11);
    return 2 if($ndreams < 21);
    return (POSIX::ceil($ndreams/10)+1) if($ndreams < 100);
    return (POSIX::ceil($ndreams/100)+1)*5;
  }


sub communicate
  {
    my $message= shift @_;
    my $buttons = ['OK','Cancel'];
    $buttons = shift @_ if(@_);
    my $dialog=$main::toplevel -> Dialog(
			      -bg => $main::dream_options{'bgcolor'},
			      -fg => $main::dream_options{'fgcolor'},
			      -text => $message,
			      -buttons => $buttons);
    return $dialog -> Show();
  }


sub communicate_filename
  {
    my $message = shift @_;
    my $types=[['All Files',   '*'],];
    $types = shift @_ if(@_);
    my $tempname;
    if($main::windows_os)
      {
	$tempname= $main::toplevel
	  ->getOpenFile(
			-title => 
			$message, 
		       );
      }
    else
      {
	$tempname= $main::toplevel
	  ->getOpenFile(
			-filetypes=>$types,
			-title => 
			$message, 
			-bg => $main::dream_options{'bgcolor'},
			-fg => $main::dream_options{'fgcolor'},);
      }
    return $tempname;
  }

sub help_overview
  {
    my $help1 = MainWindow -> new();
    my $help_text = $help1 ->
      Scrolled('ROText',
	       -fg => $main::dream_options{'dreamfg'}->[5],
	       -bg => $main::dream_options{'dreambg'}->[5],
	       -wrap => 'word',
	       -font => $main::dream_options{'main-font'},
	       -height => 20,
	       -spacing3=>'2m',
	       -scrollbars => 'w');
    my $tw=$help_text -> Subwidget('rotext');
    $tw -> insert("1.0","This software allows the user to create a diary of dreams and store it securely. It opens two windows:

An upper window for the text of a dream.

A lower window for interpretations or comments. 

Each window can contain styled (unicode) text and images. 

The lower window can contain specialized styles of text that
cause automatic searches on the dream text in the upper window.

The first time you run it, it initializes a new database.
");
$help_text -> pack();
  }


sub help_search
  {
    my $help1 = MainWindow -> new();
    my $help_text = $help1 ->
      Scrolled('ROText',
	       -fg => $main::dream_options{'dreamfg'}->[5],
	       -bg => $main::dream_options{'dreambg'}->[5],
	       -wrap => 'word',
	       -font => $main::dream_options{'main-font'},
	       -height => 20,
	       -spacing3=>'2m',
	       -scrollbars => 'w');
    my $tw=$help_text -> Subwidget('rotext');
    $tw -> insert("1.0","
1.  \'Active-Quote\' (Alt-B) highlights text in the interpretation
and makes it into a hyperlink: clicking on it with the third mouse button
causes a search in the database on that phrase.

2. \'Active-Date\' is similar but it searches for a date.

3. You can search the dream database on words or phrases. After such a
search, selecting \'Undo Search\' under the Search menu returns the database to the state it was in prior to the search.");
$help_text -> pack();
  }

sub help_insert_image
  {
    my $help1 = MainWindow -> new();
    my $help_text = $help1 ->
      Scrolled('ROText',
	       -fg => $main::dream_options{'dreamfg'}->[5],
	       -bg => $main::dream_options{'dreambg'}->[5],
	       -wrap => 'word',
	       -font => $main::dream_options{'main-font'},
	       -height => 20,
	       -spacing3=>'2m',
	       -scrollbars => 'w');
    my $tw=$help_text -> Subwidget('rotext');
    $tw -> insert("1.0","
Under the \'Images\" menu:
1.  \'Add image from file\' --- does as the name implies. The image may be any format Imagemagick recognizes.

2. \'Paste image from web\'--- will download an image and insert it into the document. You must insert the image's URL into the dialog box that appears. 

Note: Many web sites will not allow \'deep access\' to images. This is accessing an image without accessing the links that led to it. Web sites will regard your dream diary as a web browser --- and not the one you used to find the image. In this case, pasting the image from the web will have no effect. The only solution is to use your web browser to download the image into a file and select the \'Add image from file\' option.");
$help_text -> pack();
  }
